import asyncio
import uuid
from typing import List, Union

from playwright.async_api import async_playwright
from playwright_stealth import Stealth
from pydantic import BaseModel

from pyba.core.lib.action import perform_action
from pyba.core.lib.mode.base import BaseEngine
from pyba.core.scripts import LoginEngine
from pyba.database import Database
from pyba.utils.common import initial_page_setup
from pyba.utils.exceptions import PromptNotPresent, UnknownSiteChosen
from pyba.utils.load_yaml import load_config

config = load_config("general")


class Step(BaseEngine):
    """
    Step-by-step browser automation. The user controls the loop externally
    by calling start(), step(), and stop().

    Args:
        `openai_api_key`: API key for OpenAI models should you want to use that
        `vertexai_project_id`: Create a VertexAI project to use that instead of OpenAI
        `vertexai_server_location`: VertexAI server location
        `gemini_api_key`: API key for Gemini-2.5-pro native support without VertexAI
        `use_random`: Enables mouse and scroll randomisations to evade bot detection
        `headless`: Choose if you want to run in the headless mode or not
        `handle_dependencies`: Choose if you want to automatically install dependencies during runtime
        `use_logger`: Choose if you want to use the logger (that is enable logging of data)
        `enable_tracing`: Choose if you want to enable tracing. This will create a .zip file which you can use in traceviewer
        `trace_save_directory`: The directory where you want the .zip file to be saved
        `database`: An instance of the Database class which will define all database specific configs
        `get_output`: In addition to these, the step engine uses another argument to enable or disable outputs after each step
    """

    def __init__(
        self,
        openai_api_key: str = None,
        vertexai_project_id: str = None,
        vertexai_server_location: str = None,
        gemini_api_key: str = None,
        headless: bool = False,
        handle_dependencies: bool = config["main_engine_configs"]["handle_dependencies"],
        use_random: bool = config["main_engine_configs"]["use_random"],
        use_logger: bool = config["main_engine_configs"]["use_logger"],
        enable_tracing: bool = config["main_engine_configs"]["enable_tracing"],
        trace_save_directory: str = None,
        max_actions_per_step: int = 5,
        database: Database = None,
        get_output: bool = False,
    ):
        self.mode = "STEP"
        super().__init__(
            headless=headless,
            enable_tracing=enable_tracing,
            trace_save_directory=trace_save_directory,
            database=database,
            use_random=use_random,
            use_logger=use_logger,
            mode=self.mode,
            handle_dependencies=handle_dependencies,
            openai_api_key=openai_api_key,
            vertexai_project_id=vertexai_project_id,
            vertexai_server_location=vertexai_server_location,
            gemini_api_key=gemini_api_key,
        )

        self.session_id = uuid.uuid4().hex
        selectors = tuple(config["process_config"]["selectors"])
        self.combined_selector = ", ".join(selectors)
        self.max_actions_per_step = max_actions_per_step

        self._cleaned_dom = None
        self._playwright_context_manager = None
        self._pw = None
        self.get_output = get_output

    async def start(self, automated_login_sites: List[str] = None):
        """
        Creates a persistent browser instance. This needs to be explicitly called
        by the user when using the Step mode. This handles the automated login
        for us as well.
        """
        if automated_login_sites is not None:
            assert isinstance(
                automated_login_sites, list
            ), "Make sure the automated_login_sites is a list!"
            for engine in automated_login_sites:
                if hasattr(LoginEngine, engine):
                    self.automated_login_engine_classes.append(getattr(LoginEngine, engine))
                else:
                    raise UnknownSiteChosen(LoginEngine.available_engines())

        self._playwright_context_manager = Stealth().use_async(async_playwright())
        self._pw = await self._playwright_context_manager.__aenter__()
        self.browser = await self._pw.chromium.launch(headless=self.headless_mode)
        self.context = await self.get_trace_context()
        self.page = await self.context.new_page()
        self._cleaned_dom = await initial_page_setup(self.page)

    async def step(
        self, prompt_step: str, extraction_format: BaseModel = None
    ) -> Union[str, None]:
        """
        The step function is a replica of the `Engine.run()`. It takes in the previous action as well
        into context and tries to figure out the best way to achieve the short term prompt given by the user.

        Args:
            `prompt_step`: A single step wise prompt given by the user (This might require more than one steps)
            `extraction_format`: The final extraction format IF NEEDED
        """
        if prompt_step is None:
            raise PromptNotPresent()

        previous_action = None

        for _ in range(self.max_actions_per_step):
            login_attempted_successfully = await self.attempt_login()
            if login_attempted_successfully:
                self._cleaned_dom = await self.successful_login_clean_and_get_dom()
                continue

            action = self.fetch_action(
                cleaned_dom=self._cleaned_dom.to_dict(),
                user_prompt=prompt_step,
                previous_action=previous_action,
                extraction_format=extraction_format,
                fail_reason=None,
                action_status=True,
            )

            if self.get_output:
                output = await self.generate_output(
                    action=action, cleaned_dom=self._cleaned_dom, prompt=prompt_step
                )
                if output:
                    return output

            # Otherwise simply exit if no action
            if not action:
                return None

            self.log.action(action)
            value, fail_reason = await perform_action(self.page, action)

            if value is None:
                if self.db_funcs:
                    self.db_funcs.push_to_episodic_memory(
                        session_id=self.session_id,
                        action=str(action),
                        page_url=str(self.page.url),
                        action_status=False,
                        fail_reason=fail_reason,
                    )
                self._cleaned_dom = await self.extract_dom()

                output = await self.retry_perform_action(
                    cleaned_dom=self._cleaned_dom.to_dict(),
                    prompt=prompt_step,
                    previous_action=str(action),
                    action_status=False,
                    fail_reason=fail_reason,
                )
                if output:
                    return output
            else:
                if self.db_funcs:
                    self.db_funcs.push_to_episodic_memory(
                        session_id=self.session_id,
                        action=str(action),
                        page_url=str(self.page.url),
                        action_status=True,
                        fail_reason=None,
                    )

            previous_action = str(action)
            self._cleaned_dom = await self.extract_dom()

        return None

    async def stop(self):
        """
        Kills the persistent browser instance once called. For using the Step engine, this NEEDS
        to be called explicitly by the user in order to close the instance.
        """
        try:
            await self.save_trace()
            await self.shut_down()
        finally:
            if self._playwright_context_manager:
                await self._playwright_context_manager.__aexit__(None, None, None)

    # Some helper functions for sync endpoints
    # Note that using these will be a little weireder in the main pipeline.
    def sync_start(self, automated_login_sites: List[str] = None):
        asyncio.run(self.start(automated_login_sites=automated_login_sites))

    def sync_step(self, prompt_step: str, extraction_format: BaseModel = None) -> Union[str, None]:
        return asyncio.run(self.step(prompt_step=prompt_step, extraction_format=extraction_format))

    def sync_stop(self):
        asyncio.run(self.stop())



step_system_prompt = """
You are the **Brain** of a browser automation engine operating in **step-by-step mode**.

The user provides one instruction at a time. You must perform ONLY what the current instruction asks — nothing more.
Do not anticipate, infer, or execute future steps beyond the scope of the current instruction.
Once the instruction is fulfilled, immediately return None.

The browser always starts on Brave Search (https://search.brave.com). If the user says "search for X", simply type it into the existing search box and press Enter. Do not navigate elsewhere to search — use the Brave Search page you are already on.

Examples:
- "search for python" → fill the Brave search box with "python", press Enter, return None.
- "go to youtube.com" → navigate to youtube.com, return None. Do not start searching.
- "go to youtube.com and search for python" → navigate to youtube.com, search for python, return None. Do not click results.

---

## What You Receive

1. **Instruction (User Goal):**
   - A single, scoped instruction describing what to do right now.

2. **Cleaned DOM (Context of Current Page):**
   A structured dictionary containing:
   - `hyperlinks`: list of all hyperlink texts or targets.
   - `input_fields`: list of all fillable input elements.
   - `clickable_fields`: list of clickable elements.
   - `actual_text`: visible text content of the page.

---

## What You Must Output

A valid JSON object of type `PlaywrightResponse`:

```python
class PlaywrightResponse(BaseModel):
    actions: List[PlaywrightAction]
    extract_info: bool
```

## PlaywrightAction Schema

Set exactly one atomic action (or one valid pair like fill_selector + fill_value) per output.
All other fields must remain null or absent.

### Navigation
- goto, go_back, go_forward, reload

### Interactions
- click, dblclick, hover, right_click

### Input Actions
- fill_selector + fill_value
- type_selector + type_text
- press_selector + press_key
- check, uncheck
- select_selector + select_value
- upload_selector + upload_path

### Scrolling & Waiting
- scroll_x, scroll_y
- wait_selector, wait_timeout, wait_ms

### Dropdown Menus (specify both together)
- dropdown_field_id + dropdown_field_value

### Keyboard & Mouse
- keyboard_press, keyboard_type
- mouse_move_x, mouse_move_y
- mouse_click_x, mouse_click_y

### Page & Context Management
- new_page, close_page, switch_page_index

### Evaluation & Utilities
- evaluate_js, screenshot_path, download_selector

## Rules

### Scope
- Do ONLY what the current instruction asks. Nothing more.
- If the instruction says "go to youtube.com", navigate there and return None. Do not start searching.
- If the instruction says "search for X", fill the search box and press Enter. Do not click results.

### Atomicity
- Exactly one atomic PlaywrightAction per response.
- Complex operations must be split across responses.

### Contextual Validity
- Only reference elements present in the cleaned DOM.
- Do not invent or guess selectors.

### Extract Info
- Set extract_info to true if the current page contains information relevant to the user's instruction.
- Do NOT extract content yourself. Just set the flag.

### Completion
- Once the instruction is fully satisfied, return None immediately.
- Do not continue with actions beyond the instruction's scope.

NOTE: IF THE USER HAS REQUESTED FOR CERTAIN EXTRACTIONS, DON'T TRY TO DO IT YOURSELF. SET THE `extract_info` BOOLEAN TO TRUE AND PROCEED (OR SET A WAIT TIME IN ACTIONS)
"""



import asyncio
import uuid
from typing import List, Union

from playwright.async_api import async_playwright
from playwright_stealth import Stealth
from pydantic import BaseModel

from pyba.core.lib.action import perform_action
from pyba.core.lib.mode.base import BaseEngine
from pyba.core.scripts import LoginEngine
from pyba.database import Database
from pyba.utils.common import initial_page_setup
from pyba.utils.exceptions import PromptNotPresent, UnknownSiteChosen
from pyba.utils.load_yaml import load_config

config = load_config("general")


class Engine(BaseEngine):
    """
    The main entrypoint for browser automation. This engine exposes the main entry point which is the run() method

    Args:
        `openai_api_key`: API key for OpenAI models should you want to use that
        `vertexai_project_id`: Create a VertexAI project to use that instead of OpenAI
        `vertexai_server_location`: VertexAI server location
        `gemini_api_key`: API key for Gemini-2.5-pro native support without VertexAI
        `headless`: Choose if you want to run in the headless mode or not
        `handle_dependencies`: Choose if you want to automatically install dependencies during runtime
        `use_logger`: Choose if you want to use the logger (that is enable logging of data)
        `enable_tracing`: Choose if you want to enable tracing. This will create a .zip file which you can use in traceviewer
        `trace_save_directory`: The directory where you want the .zip file to be saved
        `max_depth`: The maximum number of actions that you want the model to execute
        `database`: An instance of the Database class which will define all database specific configs

    Find these default values at `pyba/config.yaml`.

    The `Engine` is inherited off from the `BaseEngine`. The BaseEngine handles the common methods for
    all the modes (default, DFS and BFS). The main `Engine` decides if execution needs to be passed to a different
    mode depending on what is set by the user.

    """

    def __init__(
        self,
        openai_api_key: str = None,
        vertexai_project_id: str = None,
        vertexai_server_location: str = None,
        gemini_api_key: str = None,
        headless: bool = config["main_engine_configs"]["headless_mode"],
        handle_dependencies: bool = config["main_engine_configs"]["handle_dependencies"],
        use_random: bool = config["main_engine_configs"]["use_random"],
        use_logger: bool = config["main_engine_configs"]["use_logger"],
        enable_tracing: bool = config["main_engine_configs"]["enable_tracing"],
        trace_save_directory: str = None,
        max_depth: int = config["main_engine_configs"]["max_iteration_steps"],
        database: Database = None,
    ):
        self.mode = "Normal"
        # Passing the common setup to the BaseEngine
        super().__init__(
            headless=headless,
            enable_tracing=enable_tracing,
            trace_save_directory=trace_save_directory,
            database=database,
            use_random=use_random,
            use_logger=use_logger,
            mode=self.mode,
            handle_dependencies=handle_dependencies,
            openai_api_key=openai_api_key,
            vertexai_project_id=vertexai_project_id,
            vertexai_server_location=vertexai_server_location,
            gemini_api_key=gemini_api_key,
        )

        self.max_depth = max_depth
        # session_id stays here becasue BaseEngine will be inherited by many
        self.session_id = uuid.uuid4().hex

        selectors = tuple(config["process_config"]["selectors"])
        self.combined_selector = ", ".join(selectors)

    async def run(
        self,
        prompt: str = None,
        automated_login_sites: List[str] = None,
        extraction_format: BaseModel = None,
    ):
        """
        The most basic implementation for the run function

        Args:
            `prompt`: The user's instructions. This is a well defined instruction.
            `automated_login_sites`: A list of sites that you want the model to automatically login to using env credentials
            `extraction_format`: A pydantic BaseModel which defines the extraction format for any data extraction

        Note:

        The `extraction_format` will be decided based on every action. For example:

        ```python3
        from pydantic import BaseModel
        from pyba import Engine

        task = "Go to hackernews. For each post, extract the title, number of upvotes and comments, and the description too"

        class Output(BaseModel):
            # Using optional is a good idea in case the things you're looking for don't exist
            title: Optional[str],
            num_upvotes: Optional[int],
            num_comments: Optional[int],
            desc: Optional[str]

        engine = Engine(**kwargs)

        await engine.run(task, extraction_format=Output)
        ```

        would return data **during** the execution, and now once it finishes. It will dump it in the database as well, and it
        decides if data needs to be extracted on an action basis.

        Using this feature will NOT cost you any more tokens than usual.
        """
        if prompt is None:
            raise PromptNotPresent()

        if automated_login_sites is not None:
            assert isinstance(
                automated_login_sites, list
            ), "Make sure the automated_login_sites is a list!"

            for engine in automated_login_sites:
                # Each engine is going to be a name like "instagram"
                if hasattr(LoginEngine, engine):
                    engine_class = getattr(LoginEngine, engine)
                    self.automated_login_engine_classes.append(engine_class)
                else:
                    raise UnknownSiteChosen(LoginEngine.available_engines())
        try:
            async with Stealth().use_async(async_playwright()) as p:
                self.browser = await p.chromium.launch(headless=self.headless_mode)

                self.context = await self.get_trace_context()
                self.page = await self.context.new_page()
                cleaned_dom = await initial_page_setup(self.page)

                for steps in range(0, self.max_depth):
                    # If LoginEngines have been chosen then self.automated_login_engine_classes will be populated
                    login_attempted_successfully = await self.attempt_login()
                    if login_attempted_successfully:
                        cleaned_dom = await self.successful_login_clean_and_get_dom()
                        # Jump to the next iteration of the loop
                        continue

                    # Get an actionable PlaywrightResponse from the models, along with `extracted results` if any

                    # NOTE: This function needs to actually fetch history, but right now its fetching the previous_action only
                    # We need to ensure that we store the previous action REGARDLESS of whether a database has been provided by
                    # the user or not

                    # history = self.fetch_history()
                    previous_action = (
                        self.fetch_history()
                    )  # This needs to change... Do not forget to change this in DFS mode as well.
                    action = self.fetch_action(
                        cleaned_dom=cleaned_dom.to_dict(),
                        user_prompt=prompt,
                        previous_action=previous_action,
                        extraction_format=extraction_format,
                        fail_reason=None,
                        action_status=True,
                    )
                    output = await self.generate_output(
                        action=action, cleaned_dom=cleaned_dom, prompt=prompt
                    )

                    if output:
                        await self.save_trace()
                        await self.shut_down()
                        return output

                    self.log.action(action)

                    # If its not None, then perform it
                    value, fail_reason = await perform_action(self.page, action)

                    if value is None:
                        # This means the action failed due to whatever reason. The best bet is to
                        # pass in the latest cleaned_dom and get the output again
                        if self.db_funcs:
                            self.db_funcs.push_to_episodic_memory(
                                session_id=self.session_id,
                                action=str(action),
                                page_url=str(self.page.url),
                                action_status=False,
                                fail_reason=fail_reason,
                            )
                        cleaned_dom = await self.extract_dom()

                        output = await self.retry_perform_action(
                            cleaned_dom=cleaned_dom.to_dict(),
                            prompt=prompt,
                            previous_action=previous_action,
                            action_status=False,  # This is pretty much unncessary because passing a fail_reason DOES mean that action failed, but it makes it a little more coherent
                            fail_reason=fail_reason,
                        )

                        if output:
                            await self.save_trace()
                            await self.shut_down()
                            return output

                    # Else, get the new DOM and restart loop
                    if self.db_funcs:
                        self.db_funcs.push_to_episodic_memory(
                            session_id=self.session_id,
                            action=str(action),
                            page_url=str(self.page.url),
                            action_status=True,  # Defaults to true anyway so I don't HAVE to do this
                            fail_reason=None,
                        )
                    cleaned_dom = await self.extract_dom()
        finally:
            await self.save_trace()
            await self.shut_down()

    def sync_run(
        self,
        prompt: str = None,
        automated_login_sites: List[str] = None,
        extraction_format: BaseModel = None,
    ) -> Union[str, None]:
        """
        Sync endpoint for running the above function
        """
        output = asyncio.run(
            self.run(
                prompt=prompt,
                automated_login_sites=automated_login_sites,
                extraction_format=extraction_format,
            )
        )

        if output:
            return output



import json
import time
from typing import Optional, List

from pyba.database.database import Database
from pyba.database.models import EpisodicMemory, SemanticMemory, BFSEpisodicMemory


class DatabaseFunctions:
    """
    Composition class for the database functions
    """

    def __init__(self, database: Database):
        """
        Args:
            `database`: The database instance for commiting

        If database is none, it doesn't initialise the database session
        """
        if database is None:
            return
        self.database = database
        self.session = self.database.session

    def submit_query_with_retry(self):
        """
        Function to send submit based queries to db
        (such as insert and update or delete), it retries 100 times if
        connection returned an error.

        Args:
            `session`: session to commit

        Returns:
            True if submitted success otherwise False
        """
        if not hasattr(self, "session"):
            return False

        try:
            for _ in range(1, 100):
                try:
                    self.session.commit()
                    return True
                except Exception:
                    time.sleep(0.1)
        except Exception:
            self.session.rollback()
            return False

        self.session.rollback()
        return False

    def push_to_episodic_memory(
        self,
        session_id: str,
        action: str,
        page_url: str,
        action_status: bool,
        fail_reason: str = None,
    ) -> bool:
        """
        Pushes a new action and page_url onto the stack for a given session_id.
        It retrieves the existing record, appends the new values as JSON strings,
        and updates/inserts the record.

        Args:
            `session_id`: The unique session ID.
            `action`: The action string to be pushed.
            `page_url`: The page URL string to be pushed.
            `action_status`: The success or the failure of the current action (True -> Success, False -> Failure)
            `fail_reason`: A string to dictate why a particular action failed (defaults to None in an event of success)

        Returns:
            True if the operation was successful, otherwise False.
        """
        if not hasattr(self, "session"):
            return False
        try:
            memory_record = (
                self.session.query(EpisodicMemory)
                .filter(EpisodicMemory.session_id == session_id)
                .one_or_none()
            )

            if memory_record:
                try:
                    actions_list = json.loads(memory_record.actions)
                    page_url_list = json.loads(memory_record.page_url)
                    action_status_list = json.loads(memory_record.action_status)
                    fail_reason_list = json.loads(memory_record.fail_reason)
                except json.JSONDecodeError:
                    # If stored data is not a valid json, refresh it with a new list
                    actions_list = []
                    page_url_list = []
                    action_status_list = []
                    fail_reason_list = []

                actions_list.append(action)
                page_url_list.append(page_url)
                action_status_list.append(action_status)
                fail_reason_list.append(fail_reason)

                memory_record.actions = json.dumps(actions_list)
                memory_record.page_url = json.dumps(page_url_list)
                memory_record.action_status = json.dumps(action_status_list)
                memory_record.fail_reason = json.dumps(fail_reason_list)

            else:
                new_memory = EpisodicMemory(
                    session_id=session_id,
                    actions=json.dumps([action]),
                    page_url=json.dumps([page_url]),
                    action_status=json.dumps([action_status]),
                    fail_reason=json.dumps([fail_reason]),
                )
                self.session.add(new_memory)

            return self.submit_query_with_retry()

        except Exception:
            self.session.rollback()
            return False
        finally:
            self.session.close()

    def get_episodic_memory_by_session_id(self, session_id: str) -> Optional[EpisodicMemory]:
        """
        Retrieves an episodic memory record by its `session_id`.

        Args:
            `session_id`: The unique session ID to query for.
        Returns:
            An EpisodicMemory object if found, else None.
        """
        # Check if session exists
        if not hasattr(self, "session"):
            return None
        try:
            memory = self.session.get(EpisodicMemory, session_id)
            return memory
        except Exception:
            return None

    def push_to_bfs_episodic_memory(
        self, session_id: str, context_id: str, action: str, page_url: str
    ) -> bool:
        """
        Pushes a new action and page_url for a specific BFS context.
        Creates a new record if the (session_id, context_id) pair doesn't exist,
        otherwise appends to the existing record.

        Note: This function uses a composite primary key of (session_id, context_id) to allow multiple
        browser windows per session.

        Args:
            `session_id`: The parent session ID for the BFS run
            `context_id`: The unique context ID for this browser window
            `action`: The action string to be pushed
            `page_url`: The page URL string to be pushed

        Returns:
            True if the operation was successful, otherwise False.
        """
        if not hasattr(self, "session"):
            return False

        try:
            memory_record = (
                self.session.query(BFSEpisodicMemory)
                .filter(
                    BFSEpisodicMemory.session_id == session_id,
                    BFSEpisodicMemory.context_id == context_id,
                )
                .one_or_none()
            )

            if memory_record:
                try:
                    actions_list = json.loads(memory_record.actions)
                    page_url_list = json.loads(memory_record.page_url)
                except json.JSONDecodeError:
                    actions_list = []
                    page_url_list = []

                actions_list.append(action)
                page_url_list.append(page_url)

                memory_record.actions = json.dumps(actions_list)
                memory_record.page_url = json.dumps(page_url_list)

            else:
                new_memory = BFSEpisodicMemory(
                    session_id=session_id,
                    context_id=context_id,
                    actions=json.dumps([action]),
                    page_url=json.dumps([page_url]),
                )
                self.session.add(new_memory)

            return self.submit_query_with_retry()

        except Exception:
            self.session.rollback()
            return False
        finally:
            self.session.close()

    def get_bfs_episodic_memory_by_context(
        self, session_id: str, context_id: str
    ) -> Optional[BFSEpisodicMemory]:
        """
        Retrieves a specific BFS context's episodic memory. Needs both the session_id and the context_id to retrieve the correct record.

        Args:
            `session_id`: The parent session ID
            `context_id`: The specific context ID to retrieve

        Returns:
            A BFSEpisodicMemory object if found, else None
        """
        if not hasattr(self, "session"):
            return None
        try:
            memory = (
                self.session.query(BFSEpisodicMemory)
                .filter(
                    BFSEpisodicMemory.session_id == session_id,
                    BFSEpisodicMemory.context_id == context_id,
                )
                .one_or_none()
            )
            return memory
        except Exception:
            return None

    def get_all_bfs_contexts_by_session(
        self, session_id: str
    ) -> Optional[List[BFSEpisodicMemory]]:
        """
        Retrieves all BFS context records for a given session.

        Args:
            `session_id`: The parent session ID to query for

        Returns:
            A list of BFSEpisodicMemory objects for all contexts in the session,
            or None if no records found or error occurred.
        """
        if not hasattr(self, "session"):
            return None
        try:
            memories = (
                self.session.query(BFSEpisodicMemory)
                .filter(BFSEpisodicMemory.session_id == session_id)
                .all()
            )
            return memories if memories else None
        except Exception:
            return None

    def push_to_semantic_memory(self, session_id: str, logs: str) -> bool:
        """
        Pushes logs to semantic memory.

        Args:
            `session_id`: The unique session ID
            `logs`: A dump generated by the memory generator

        Returns:
            A boolean to indicate the success or failure of the operation
        """
        print("in here")
        if not hasattr(self, "session"):
            return False

        try:
            # Access the current memory record if it exists
            memory_record = (
                self.session.query(SemanticMemory)
                .filter(SemanticMemory.session_id == session_id)
                .one_or_none()
            )

            if memory_record:
                try:
                    logs_list = json.loads(memory_record.logs)
                except json.JSONDecodeError:
                    logs_list = []

                logs_list.append(logs)

                memory_record.logs = json.dumps(logs_list)

            else:
                new_memory = SemanticMemory(
                    session_id=session_id,
                    logs=json.dumps([logs]),
                )
                self.session.add(new_memory)

            return self.submit_query_with_retry()

        except Exception as e:
            print(f"Hit exception: {e}")
            self.session.rollback()
            return False
        finally:
            self.session.close()

    def get_semantic_memory_by_session_id(self, session_id: str) -> Optional[SemanticMemory]:
        """
        Retrieves all the semantic memory from the database

        Args:
            `session_id`: The unique session ID to query for.
        Returns:
            An EpisodicMemory object if found, else None.
        """
        if not hasattr(self, "session"):
            return None
        try:
            memory = self.session.get(SemanticMemory, session_id)
            return memory
        except Exception:
            return None



general_prompt = """
You are the Brain of a browser-automation engine.

You operate in a strict step-by-step loop.
At each step, you must choose **exactly one atomic PlaywrightAction** that moves the task closer to completion.

You can only see the webpage through the structured DOM snapshot provided.
You must reason **only** from this data.
Do not assume, infer, or hallucinate anything not explicitly present.

Your output must strictly follow the rules below.

# Instructions

## Core responsibilities
- Decide exactly one atomic PlaywrightAction.
- Decide whether the current page contains information that should be extracted.
- Never perform extraction yourself; only signal it.

## Atomicity rules
- Exactly one PlaywrightAction per step.
- Only one actionable field may be non-null.
- Selector–value pairs count as a single field:
  - fill_selector + fill_value
  - type_selector + type_text
  - press_selector + press_key
  - select_selector + select_value
  - upload_selector + upload_path
- All other action fields must be null or omitted.

## Action constraints
- Never merge steps.
- Typing then pressing Enter = two steps.
- Filling then clicking = two steps.

## Selector constraints
- Selectors must appear **verbatim** in the provided DOM snapshot.
- No guessing, no generalisation, no invention.

## Goal progression
- Always choose the smallest logical action that advances the user’s goal.
- If you just filled an input, the next step is usually pressing Enter on that same selector.
- If nothing obvious matches the goal, press Enter on the most relevant input field.

## Extraction logic
- Output a boolean `extract_info`.
- Set `extract_info = true` if the current page visibly contains information required by the user goal.
- Do NOT extract or summarise content yourself.
- If extraction is required before continuing, wait using a wait-type action.

## Completion
- If the task is finished and no further actions are required, return `None`.

## Output format
- Respond **only** with a valid JSON object of type `PlaywrightResponse`.

### Valid example

{{
  "actions": [
    {{
      "fill_selector": "input[name='q']",
      "fill_value": "python"
    }}
  ],
  "extract_info": false
}}

### Valid follow-up example

{{
  "actions": [
    {{
      "press_selector": "input[name='q']",
      "press_key": "Enter"
    }}
  ],
  "extract_info": false
}}

### Invalid example (multiple active actions)

{{
  "actions": [
    {{
      "click_selector": "#btn",
      "fill_selector": "#search",
      "fill_value": "hi"
    }}
  ],
  "extract_info": false
}}

# This is the new runtime data

## USER GOAL
{user_prompt}

## CURRENT PAGE CONTEXT

Current URL:
{current_url}

Hyperlinks:
{hyperlinks}

Input Fields:
{input_fields}

Clickable Elements:
{clickable_fields}

Visible Text:
{actual_text}

## PREVIOUS STEP CONTEXT

Previous Action:
{previous_action}

Action Status:
{action_status}

Failure Reason (only meaningful if Action Status is False, otherwise None):
{fail_reason}
"""



import json
import threading

from pydantic import BaseModel

from pyba.core.agent.base_agent import BaseAgent
from pyba.utils.prompts.extraction_prompts import extraction_general_instruction


class ExtractionAgent(BaseAgent):
    """
    This is a helper agent in all aspects. To use this, all other agents
    need to import and initialise this.

    This agent allows for threaded infomation extraction to not hinder the main pipeline flow.

    Args:
        `extraction_format`: The format which should be fitted for the extraction
    """

    def __init__(self, engine, extraction_format: BaseModel):
        super().__init__(engine=engine)  # Initialising the base params from BaseAgent

        self.extraction_format = extraction_format
        self.agent = self.llm_factory.get_extraction_agent(
            extraction_format=self.extraction_format
        )  # Getting the extraction agent

    def _initialise_prompt(self, task: str, actual_text: str):
        """
        Takes in the actual_text and wraps it around the general prompt

        Args:
                `task`: The user's defined task
                `actual_text`: The current text on the page
        """
        return extraction_general_instruction.format(task=task, actual_text=actual_text)

    def info_extraction(self, task: str, actual_text: str, context_id: str = None) -> None:
        """
        Function to extract data from the current page

        Args:
            `task`: The user's defined task
            `actual_text`: The current page text
            `context_id`: A unique identifier for this browser window (useful when multiple windows)

        This function for now only Logs the value and doesn't return anything
        """

        # THE FINAL PIECE OF THE PUZZLE
        prompt = self._initialise_prompt(task=task, actual_text=actual_text)

        if self.engine.provider == "openai":
            response = self.handle_openai_execution(
                agent=self.agent,
                prompt=prompt,
                context_id=context_id,
            )
            try:
                parsed_json = json.loads(response.choices[0].message.content)
                self.log.info(f"Extracted content: {parsed_json}")
                if self.engine.db_funcs:
                    self.engine.db_funcs.push_to_semantic_memory(
                        self.engine.session_id, logs=json.dumps(parsed_json)
                    )
                    self.log.info("Added to semantic memory")
            except Exception as e:
                self.log.error(f"Unable to parse the outoput from OpenAI response: {e}")
                return None
        elif self.engine.provider == "vertexai":
            print("In here")
            response = self.handle_vertexai_execution(
                agent=self.agent, prompt=prompt, context_id=context_id
            )

            try:
                parsed_object = getattr(
                    response, "output_parsed", getattr(response, "parsed", None)
                )

                if not parsed_object:
                    self.log.error("No parsed object found in VertexAI response.")
                    return None

                self.log.info(f"Extracted content: {parsed_object}")
                if self.engine.db_funcs:
                    print("in here")
                    self.engine.db_funcs.push_to_semantic_memory(
                        self.engine.session_id, logs=parsed_object.json()
                    )
                    self.log.info("Added to semantic memory")

            except Exception as e:
                print(f"hit exception: {e}")
                if not response:
                    self.log.error(f"Unable to parse the output from VertexAI response: {e}")
                # If we have a response which cannot be parsed, it MUST be a None value
        else:  # Using gemini
            response = self.handle_gemini_execution(
                agent=self.agent, prompt=prompt, context_id=context_id
            )
            parsed_object = self.agent["response_format"].model_validate_json(response.text)
            self.log.info(f"Extracted content: {parsed_object}")
            if self.engine.db_funcs:
                self.engine.db_funcs.push_to_semantic_memory(
                    self.engine.session_id, logs=parsed_object.json()
                )
                self.log.info("Added to semantic memory")

    def run_threaded_info_extraction(self, task: str, actual_text: str):
        """
                Fuction to thread the execution of the `info_extraction` function

                Args:
            `task`: The user's defined task
            `actual_text`: The current page text

        This function creates a separate thread for calling the agent on the current page
        and extracting the relevant information with the right format.
        """
        self.log.info("Running the extractor on the current page")
        thread = threading.Thread(
            target=self.info_extraction,
            args=(task, actual_text),
            daemon=True,
        )
        thread.start()



import os
import time
import uuid
from typing import Literal, List

from pinecone import ServerlessSpec
from pinecone.grpc import PineconeGRPC as Pinecone


class Memory:
    def __init__(self, memory_type: Literal["pyba-sm-mem"]):
        self.memory_type = memory_type

        self.pinecone_api_key = os.getenv("pinecone_api_key")
        self.pc = Pinecone(api_key=self.pinecone_api_key)
        self.index = self.create_index(self.memory_type)

        # We use a model hosted by Pinecone.
        # 'multilingual-e5-large' uses 1024 dimensions so we don't have to change the dimensions
        self.model_name = "multilingual-e5-large"
        self.session_id = uuid.uuid4().hex

    def create_index(self, memory_type: str):
        """
        helper function to create indices for each memory structure
        """
        # if not already created
        if self.memory_type not in [idx["name"] for idx in self.pc.list_indexes()]:
            self.pc.create_index(
                name=self.memory_type,
                dimension=1024,
                metric="cosine",
                spec=ServerlessSpec(cloud="aws", region="us-east-1"),
            )

        while not self.pc.describe_index(self.memory_type).status["ready"]:
            time.sleep(1)

        return self.pc.Index(self.memory_type)

    def embed(self, text: str) -> List:
        """
        Creates an embedding for a given text.
        """
        # Pinecone expects a list of strings, we send one.
        res = self.pc.inference.embed(
            model=self.model_name, inputs=[text], parameters={"input_type": "passage"}
        )
        return res[0].values

    def save(
        self,
        text: str,
        role: Literal["user", "azno"],
        turn_id: int,
    ) -> bool:
        """
        Save a single conversational unit to Pinecone.

        Args:
            `text`: The actual text to be saved
            `role`: A role assignment for the text
            `turn_id`: A unique ID for the current turn of chats

        Usage:
        ```python3
            turn_id = uuid.uuid4().hex
            await self.mem.save(user_text, role="user", turn_id=turn_id)
            await self.mem.save(azno_text, role="azno", turn_id=turn_id)
        ```
        """

        embedding = self.embed(text)
        vector_id = f"{self.session_id}-{turn_id}-{role}"

        vector = {
            "id": vector_id,
            "values": embedding,
            "metadata": {
                "role": role,
                "turn_id": turn_id,
                "preview": text[:300],  # short, safe metadata
            },
        }

        try:
            self.index.upsert(
                vectors=[vector],
                namespace=self.session_id,
            )
            return True

        except Exception as e:
            print(f"Failed to insert vector: {e}")
            return False

    def query(self, query_text: str, top_k: int = 5) -> list:
        """
        Retrieve semantically similar past conversation units
        within the same session.

        Args:
            `query_text`: The text being used to query the DB
            `top_k`: Number of matches to return

        Usage:
            - Combine the last two messages of the user INCLUDING the current one as the `query_text`
        """

        query_embedding = self.embed(query_text)

        try:
            res = self.index.query(
                vector=query_embedding,
                top_k=top_k,
                include_metadata=True,
                namespace=self.session_id,
            )
            results = []
            for match in res.get("matches", []):
                results.append(
                    {
                        "id": match["id"],
                        "score": match["score"],
                        "role": match["metadata"]["role"],
                        "turn_id": match["metadata"]["turn_id"],
                        "preview": match["metadata"]["preview"],
                    }
                )

            return results

        except Exception as e:
            print(f"Query failed: {e}")
            return []


# Future implementation ideas for the vector DB
# Store the raw logs in a SQL based database. Then every ~10-20 turns, store the ENTIRE SUMMARISED
# history in the database.
# We'll see about that, I also need to make the layer on top of this.

# if __name__ == "__main__":
#     mem = Memory(memory_type="azno-ep-mem")
#     a = mem.query("Hello")
#     print(a)



import random
import time
from typing import Literal, Dict, List, Any

from pyba.core.agent.llm_factory import LLMFactory
from pyba.logger import get_logger


class BaseAgent:
    """
    The base class for all Agents to define common methods

        Contains methods for exponential backoff and retry as well
        Note: this backoff and retry will be blocking for that specific context.

    Defines the following variables:

    `exponential_base`: 2 (we're using base 2)
    `base_timeout`: 1 second
    `max_backoff_time`: 60 seconds
    `attempt_number`: The current attempt number initialised to 1
    `LLMFactory`: The internal agent call is made by agent itself
    `log`: The logger for the agents
    """

    def __init__(self, engine):
        self.base = 2
        self.base_timeout = 1
        self.max_backoff_time = 60

        self.engine = engine
        self.llm_factory = LLMFactory(engine=self.engine)
        self.log = get_logger()
        self.mode: Literal["Normal", "DFS", "BFS"] = self.engine.mode
        self.shared_depth_dictionary = {}

    def _initialise_prompt(self):
        """
        Function to initialise prompts. This function needs to be impemented for each agent
        """
        raise NotImplementedError("Subclasses must implement _initialise_prompt")

    def _initialise_openai_arguments(
        self, system_instruction: str, prompt: str, model_name: str
    ) -> Dict[str, List[Dict[str, str]]]:
        """
        Initialises the arguments for OpenAI agents

        Args:
            `system_instruction`: The system instruction for the agent
            `prompt`: The current prompt for the agent
            `model_name`: The OpenAI model name

        Returns:
            An arguments dictionary which can be directly passed to OpenAI agents
        """

        messages = [
            {"role": "system", "content": system_instruction},
            {"role": "user", "content": prompt},
        ]

        kwargs = {
            "model": model_name,
            "messages": messages,
        }

        return kwargs

    def handle_openai_execution(self, agent: Any, prompt: str, context_id: str = None):
        """
        Helper method to handle OpenAI execution

        Args:
            `agent`: The agent to use (action_agent or output_agent)
            `prompt`: The fully formatted prompt string
            `context_id`: A unique identifier for the current browser window

        The `context_id` is to help in differentiating between different browser windows during parallel execution
        for BFS mode.

        `context_id`=None => There is only one browser session.

        Returns:
            `response`: The raw response from the model. The exact required values
            are expected to be extraced within each agent
        """
        arguments = self._initialise_openai_arguments(
            system_instruction=agent["system_instruction"],
            prompt=prompt,
            model_name=agent["model"],
        )

        while True:
            try:
                response = agent["client"].chat.completions.parse(
                    **arguments, response_format=agent["response_format"]
                )
                # self.attempt_number = 1
                self.initialise_depth_ladder(unique_context_id=context_id)
                break
            except Exception:
                # If we hit a rate limit, calculate the time to wait and retry
                wait_time = self.calculate_next_time(
                    self.shared_depth_dictionary.get(context_id, 1)
                )
                self.log.warning(f"Hit the rate limit for OpenAI, retrying in {wait_time} seconds")
                time.sleep(wait_time)  # wait_time is in seconds
                # self.attempt_number += 1
                self.update_depth_ladder(unique_context_id=context_id)

        return response

    def handle_vertexai_execution(self, agent: Any, prompt: str, context_id: str = None):
        """
        Helper method to handle VertexAI execution

        Args:
            `agent`: The agent to use (action_agent or output_agent)
            `prompt`: The fully formatted prompt string
            `context_id`: A unique identifier for the current browser window

        The `context_id` is to help in differentiating between different browser windows during parallel execution
        for BFS mode.

        `context_id`=None => There is only one browser session.


        Returns:
            `response`: The raw response from the model. The exact required values
            are expected to be extraced within each agent
        """
        while True:
            try:
                response = agent.send_message(prompt)
                self.initialise_depth_ladder(unique_context_id=context_id)
                break
            except Exception:
                wait_time = self.calculate_next_time(
                    self.shared_depth_dictionary.get(context_id, 1)
                )
                self.log.warning(
                    f"Hit the rate limit for VertexAI, retrying in {wait_time} seconds"
                )
                time.sleep(wait_time)
                self.update_depth_ladder(unique_context_id=context_id)
        return response

    def handle_gemini_execution(self, agent: Any, prompt: str, context_id: str = None):
        """
        Helper method to handle gemini's execution

        Args:
            `agent`: The agent to use (action_agent or output_agent)
            `prompt`: The fully formatted prompt string
            `context_id`: A unique identifier for the current browser window

        The `context_id` is to help in differentiating between different browser windows during parallel execution
        for BFS mode.

        `context_id`=None => There is only one browser session.

        Returns:
            `response`: The raw response from the model. The exact required values
            are expected to be extraced within each agent
        """
        gemini_config = {
            "response_mime_type": "application/json",
            "response_json_schema": agent["response_format"].model_json_schema(),
            "system_instruction": agent["system_instruction"],
        }

        while True:
            try:
                response = agent["client"].models.generate_content(
                    model=agent["model"],
                    contents=prompt,
                    config=gemini_config,
                )
                self.initialise_depth_ladder(unique_context_id=context_id)
                break
            except Exception:
                wait_time = self.calculate_next_time(
                    self.shared_depth_dictionary.get(context_id, 1)
                )
                self.log.warning(f"Hit the rate limit for Gemini, retrying in {wait_time} seconds")
                time.sleep(wait_time)
                self.update_depth_ladder(unique_context_id=context_id)

        return response

    def calculate_next_time(self, attempt_number):
        """
        Function to calculate the next wait time in seconds

        Args:
                `attempt_number`: The number of failed attempts
        """

        delay = self.base_timeout * (self.base ** (attempt_number - 1))
        delay = min(delay, self.max_backoff_time)
        jitter = random.uniform(0, delay / 2)
        return delay + jitter

    def initialise_depth_ladder(self, unique_context_id: str):
        """
        Initialises and helps manage the depth-ladder for different browser sessions

        Args:
            `unique_context_id`: The context ID for the current browser session
        """
        self.shared_depth_dictionary[unique_context_id] = 1

    def update_depth_ladder(self, unique_context_id: str):
        """
        This function helps increments the depth-value for each browser

        Args:
            `unique_context_id`: The context ID for the browser
        """
        self.shared_depth_dictionary[unique_context_id] += 1



import json
from types import SimpleNamespace
from typing import Dict, List, Union, Any

from pydantic import BaseModel

from pyba.core.agent.base_agent import BaseAgent
from pyba.core.agent.extraction_agent import ExtractionAgent
from pyba.utils.prompts import general_prompt, output_prompt
from pyba.utils.structure import PlaywrightResponse


class PlaywrightAgent(BaseAgent):
    """
    Defines the playwright agent's actions

    Provides two endpoints:
        - `process_action`: for returning the right action on a page
        - `get_output`: for summarizing the chat and returning a string
    """

    def __init__(self, engine) -> None:
        """
        Args:
            `engine`: holds all the arguments from the user including the mode
        """
        super().__init__(engine=engine)  # Initialising the base params from BaseAgent
        self.action_agent, self.output_agent = self.llm_factory.get_agent()

    def _initialise_prompt(
        self,
        cleaned_dom: Dict[str, Union[List, str]],
        user_prompt: str,
        main_instruction: str,
        previous_action: str = None,
        fail_reason: str = None,
        action_status: bool = None,
    ):
        """
        Method to initailise the main instruction for any agent

        Args:
            `cleaned_dom`: A dictionary containing nicely formatted DOM elements
            `user_prompt`: The instructions given by the user
            `main_instruction`: The prompt for the playwright agent
            `previous_action`: The previous action
            `fail_reason`: The reason for the failure of the previous action
            `action_status`: Boolean to decide if the previous action was a success or not

        TODO: Add `history` of ALL/SOME actions to give some context as to where we are headed

        # DEPRECATED - The fail_reason decides if the previous access was a success or not.

        For each run, a prompt containing the previous action, its status (success or failure) and a fail reason (if
        it failed) is provided. This helps the model reason better
        """

        # Adding the user_prompt to the DOM to make it easier to format the prompt
        cleaned_dom["user_prompt"] = user_prompt
        cleaned_dom["previous_action"] = previous_action
        cleaned_dom["action_status"] = action_status
        cleaned_dom["fail_reason"] = fail_reason

        prompt = main_instruction.format(**cleaned_dom)

        return prompt

    def _call_model(
        self,
        agent: Any,
        prompt: str,
        agent_type: str,
        cleaned_dom: Dict = None,
        context_id: str = None,
    ) -> Any:
        """
        Generic method to call the correct LLM provider and parse the response.

        Args:
            `agent`: The agent to use (action_agent or output_agent)
            `prompt`: The fully formatted prompt string
            `agent_type`: "action" or "output", to determine parsing logic
            `cleaned_dom`: A dictionary that holds the `actual_text` from which the data is to be extracted
            `context_id`: A unique identifier for this browser window (useful when multiple windows)

        Returns:
            The parsed response (SimpleNamespace for action, str for output)
        """

        # If this guy gives me an output which says I need to extract the relevant data from this page,
        # Then I call the extraction agent here and extract information in a separate thread? Separate thread is easier,
        # I don't have to write my functions as async then

        if self.engine.provider == "openai":
            response = self.handle_openai_execution(
                agent=agent, prompt=prompt, context_id=context_id
            )
            parsed_json = json.loads(response.choices[0].message.content)

            # Parse based on agent type
            if agent_type == "action":
                actions = SimpleNamespace(**parsed_json.get("actions")[0])
                extract_info_flag = parsed_json.get("extract_info")
                if extract_info_flag:
                    self.extractor.run_threaded_info_extraction(
                        task=self.user_prompt, actual_text=cleaned_dom["actual_text"]
                    )
                return actions
            elif agent_type == "output":
                return str(parsed_json.get("output"))

        elif self.engine.provider == "vertexai":  # VertexAI logic
            response = self.handle_vertexai_execution(
                agent=agent, prompt=prompt, context_id=context_id
            )
            try:
                parsed_object = getattr(
                    response, "output_parsed", getattr(response, "parsed", None)
                )

                if not parsed_object:
                    self.log.error("No parsed object found in VertexAI response.")
                    return None

                # Parse based on agent type
                if agent_type == "action":
                    if hasattr(parsed_object, "actions") and parsed_object.actions:
                        actions = parsed_object.actions[0]
                        extract_info_flag = parsed_object.extract_info
                        if extract_info_flag:
                            self.extractor.run_threaded_info_extraction(
                                task=self.user_prompt, actual_text=cleaned_dom["actual_text"]
                            )
                        return actions
                    raise IndexError("No 'actions' found in VertexAI response.")
                elif agent_type == "output":
                    if hasattr(parsed_object, "output") and parsed_object.output:
                        return str(parsed_object.output)
                    raise IndexError("No 'output' found in VertexAI response.")

            except Exception as e:
                if not response:
                    self.log.error(f"Unable to parse the output from VertexAI response: {e}")
                # If we have a response which cannot be parsed, it MUST be a None value

        else:  # Using gemini
            response = self.handle_gemini_execution(
                agent=agent, prompt=prompt, context_id=context_id
            )
            parsed_object = agent["response_format"].model_validate_json(response.text)
            if agent_type == "action":
                actions = parsed_object.actions[0]
                extract_info_flag = parsed_object.extract_info
                if extract_info_flag:
                    self.extractor.run_threaded_info_extraction(
                        task=self.user_prompt, actual_text=cleaned_dom["actual_text"]
                    )
                return actions
            elif agent_type == "output":
                return str(parsed_object.output)

    def process_action(
        self,
        cleaned_dom: Dict[str, Union[List, str]],
        user_prompt: str,
        previous_action: str = None,
        fail_reason: str = None,
        extraction_format: BaseModel = None,
        context_id: str = None,
        action_status: bool = None,
    ) -> PlaywrightResponse:
        """
        Method to process the DOM and provide an actionable playwright response

        Args:
            `cleaned_dom`: Dictionary of the extracted items from the DOM
                - `hyperlinks`: List
                - `input_fields` (basically all fillable boxes): List
                - `clickable_fields`: List
                - `actual_text`: string
            `user_prompt`: The instructions given by the user
            `previous_action`: The previous executed action
            `fail_reason`: Holds the fail-reason should the previous task fail
            `extraction_format`: The extraction format for the task
            `context_id`: A unique identifier for this browser window (useful when multiple windows)
            `fail_reason`: The reason for failure of the previous action (None if not provided => Action passed)
            `action_status`: The success or the failure of an action

        output:
            A predefined pydantic model called `PlaywrightResponse` which defines our DSL
        """

        prompt = self._initialise_prompt(
            cleaned_dom=cleaned_dom,
            user_prompt=user_prompt,
            main_instruction=general_prompt,
            previous_action=previous_action if previous_action else "",
            fail_reason=fail_reason if fail_reason else "",
            action_status=action_status if action_status else "",
        )

        self.user_prompt = user_prompt
        self.extractor = ExtractionAgent(engine=self.engine, extraction_format=extraction_format)

        return self._call_model(
            agent=self.action_agent,
            prompt=prompt,
            agent_type="action",
            cleaned_dom=cleaned_dom,
            context_id=context_id,
        )

    def get_output(
        self, cleaned_dom: Dict[str, Union[List, str]], user_prompt: str, context_id: str = None
    ) -> str:
        """
        Method to get the final output from the model if the user requested for one
        """

        prompt = self._initialise_prompt(
            cleaned_dom=cleaned_dom, user_prompt=user_prompt, main_instruction=output_prompt
        )

        return self._call_model(
            agent=self.output_agent, prompt=prompt, agent_type="output", context_id=context_id
        )



from sqlalchemy import Column, Text
from sqlalchemy.orm import declarative_base

Base = declarative_base()


class EpisodicMemory(Base):
    """
    Memory for history logs

    Arguments:
            - `session_id`: A unique session ID for the run
            - `actions`: A JSON string array of actions given as output by the model
            - `page_url`: A JSON string array of URLs where actions were performed
            - `action_status`: A JSON string array of booleans indicating success/failure for each action
            - `fail_reason`: A JSON string array of failure reasons (null for successful actions)

    All array fields (actions, page_url, action_status, fail_reason) are parallel arrays -
    the i-th element in each array corresponds to the same action.
    """

    __tablename__ = "EpisodicMemory"

    session_id = Column(Text, primary_key=True)
    actions = Column(Text, nullable=False)
    page_url = Column(Text, nullable=False)
    action_status = Column(
        Text, nullable=False
    )  # JSON array of booleans, e.g., '[true, false, true]'
    fail_reason = Column(
        Text, nullable=True
    )  # JSON array of strings/nulls, e.g., '[null, "timeout", null]'

    def __repr__(self):
        return (
            "EpisodicMemory(session_id: {0}, actions: {1}, page_url: {2}, action_statuses: {3}, fail_reason: {4})"
        ).format(
            self.session_id,
            self.actions,
            self.page_url,
            self.action_status,
            self.fail_reason,
        )


class SemanticMemory(Base):
    """
    Memory for holding intermediate data, relevant goals, extracted outputs

    Arguments:
            - `session_id`: A unique session ID for the run
            - `logs`: The actual logs implemented as a growing buffer

    This is a growing memory type for each session, it holds everything of relevance to the task. This memory
    will be used to summarise the final output type.

    TODO: Update this function for BFS support
    """

    __tablename__ = "SemanticMemory"

    session_id = Column(Text, primary_key=True)
    logs = Column(Text, nullable=False)

    def __repr__(self):
        return ("ExtractedData(session_id: {0}, logs: {1})").format(self.session_id, self.logs)


class BFSEpisodicMemory(Base):
    """
    Memory table specifically for BFS (Breadth-First Search) mode browser sessions.

    This uses a composite primary key of (session_id, context_id) to allow multiple
    browser contexts per session.
    Arguments:
        - `session_id`: The parent session ID that spawned these BFS contexts
        - `context_id`: A unique ID for each browser context within the session
        - `actions`: A JSON string of actions performed in this context
        - `page_url`: A JSON string of URLs visited in this context

    Each (session_id, context_id) pair represents a unique browser context's
    event log within a BFS exploration session.
    """

    __tablename__ = "BFSEpisodicMemory"

    # Composite primary key: allows multiple contexts per session
    session_id = Column(Text, primary_key=True)
    context_id = Column(Text, primary_key=True)
    actions = Column(Text, nullable=False)
    page_url = Column(Text, nullable=False)

    def __repr__(self):
        return (
            "BFSEpisodicMemory(session_id: {0}, context_id: {1}, " "actions: {2}, page_url: {3})"
        ).format(self.session_id, self.context_id, self.actions, self.page_url)



import asyncio
import random
from typing import List

from oxymouse import OxyMouse
from playwright.async_api import Page


class MouseMovements:
    """
    Class to handle random mouse movements in pyba

    These functions need to be happening async to actions such as waiting for page to load. This class
    will replace all `time.sleep()`, `asyncio.sleep()` and `wait_for_*` functionality.

    The movements are deliberately constrained to a specific region in the viewport.

    For more information on the algorithms, check out https://github.com/oxylabs/OxyMouse
    """

    def __init__(self, page: Page, width: int = 1200, height: int = 1024):
        """
        Args:
            `page`: The current page object
            `width`: The viewport width for the session, defaults at 1200
            `height`: The viewport height for the session, defaults at 1024
        """
        self.page = page

        # Picks a random number between 0 and the viewport values with mode as 500
        self.width = int(random.triangular(0, width, 600))
        self.height = int(random.triangular(0, height, 512))

    async def _run(self, algorithm: str):
        """
        Runs the mouse movements for a specified algorithm

        (NOTE: This is constrained to a small window)
        """
        mouse = OxyMouse(algorithm=algorithm)
        movements = mouse.generate_random_coordinates(
            viewport_width=self.width,
            viewport_height=self.height,
        )

        for x, y in movements:
            await self.page.mouse.move(int(x), int(y), steps=10)
            await asyncio.sleep(random.uniform(0.004, 0.015))

    async def bezier_movements(self):
        """
        Function to perform bezier curve movements
        """
        await self._run(algorithm="bezier")

    async def gaussian_movements(self):
        """
        Function to perform gaussian movements
        """
        await self._run(algorithm="gaussian")

    async def perlin_movements(self):
        """
        Function to perform perlin movements
        """

        await self._run(algorithm="perlin")

    async def random_movement(self):
        """
        Chooses a function at random
        """

        mapping = {
            1: self.bezier_movements,
            2: self.gaussian_movements,
            3: self.perlin_movements,
        }

        _ = mapping[random.randint(1, 3)]
        await _()


class ScrollMovements:
    """
    Class to mimic realistic scroll movements
    """

    def __init__(self, page: Page):
        self.page = page

    def generate_scroll_values(
        self,
        num_steps: int = 15,
        max_delta: int = 100,
        min_delta: int = 15,
    ) -> List[int]:
        """
        Helper function to generate random scroll values that sum up to zero.

        Returns a list with scroll values in pixels, positive being downscroll
        and negative being an upscroll. The final sum should be 0.
        """

        # An odd number of steps>=3 is good for jitters
        if num_steps <= 1 or num_steps % 2 != 1:
            num_steps = 15

        scroll_values: List[int] = []
        current_sum = 0

        # Generate N-1 steps randomly
        # We constrain the steps to an odd number (N-1 is even) to ensure
        # we have a balanced number of positive/negative steps, which helps
        # keep the final step small and the list sum close to zero.
        # Total number of pairs (down/up movements)
        num_pairs = (num_steps - 1) // 2

        # Generate positive (down) movements
        for _ in range(num_pairs):
            delta = random.randint(min_delta, max_delta)
            scroll_values.append(delta)
            current_sum += delta

        # Generate negative (up) movements
        for _ in range(num_pairs):
            delta = random.randint(-max_delta, -min_delta)
            scroll_values.append(delta)
            current_sum += delta

        # Adding the final step to balance the list to zero
        # This is the "jitter" that returns the scroll position to the start point
        final_delta = -current_sum
        scroll_values.append(final_delta)

        # Randomly shuffling the scroll values again
        random.shuffle(scroll_values)

        return scroll_values

    async def apply_scroll_jitters(
        self,
        num_steps: int = 15,
        max_delta: int = 100,
        min_delta: int = 15,
    ):
        """
        Performs the scroll jitters using the generated values.
        """
        scroll_values = self.generate_scroll_values(
            num_steps=num_steps,
            max_delta=max_delta,
            min_delta=min_delta,
        )

        for delta in scroll_values:
            await self.page.mouse.wheel(0, delta)
            # Adding a random delay in between
            await asyncio.sleep(random.uniform(0.05, 0.2))



import asyncio
import uuid
from typing import List, Union

from playwright.async_api import async_playwright
from playwright_stealth import Stealth
from pydantic import BaseModel

from pyba.core.agent import PlannerAgent
from pyba.core.lib.action import perform_action
from pyba.core.lib.mode.base import BaseEngine
from pyba.core.scripts import LoginEngine
from pyba.database import Database
from pyba.utils.common import initial_page_setup
from pyba.utils.exceptions import UnknownSiteChosen
from pyba.utils.load_yaml import load_config

config = load_config("general")


class DFS(BaseEngine):
    """
    Methods for handling DFS exploratory searches. The `BaseEngine` initialises
    the provider and with that the playwright action and output agents.

    This is another entry point engine and can be directly imported by the user.

    The following params are defined:

    Args:
        `openai_api_key`: API key for OpenAI models should you want to use that
        `vertexai_project_id`: Create a VertexAI project to use that instead of OpenAI
        `vertexai_server_location`: VertexAI server location
        `gemini_api_key`: API key for Gemini-2.5-pro native support without VertexAI
        `headless`: Choose if you want to run in the headless mode or not
        `handle_dependencies`: Choose if you want to automatically install dependencies during runtime
        `use_logger`: Choose if you want to use the logger (that is enable logging of data)
        `max_depth`: The maximum depth to go into for each plan, where each level of depth corresponds to an action
        `max_breadth`: The number of plans to execute one by one in depth
        `enable_tracing`: Choose if you want to enable tracing. This will create a .zip file which you can use in traceviewer
        `trace_save_directory`: The directory where you want the .zip file to be saved

        `database`: An instance of the Database class which will define all database specific configs

    Find these default values at `pyba/config.yaml`.
    """

    def __init__(
        self,
        openai_api_key: str = None,
        vertexai_project_id: str = None,
        vertexai_server_location: str = None,
        gemini_api_key: str = None,
        headless: bool = config["main_engine_configs"]["headless_mode"],
        handle_dependencies: bool = config["main_engine_configs"]["handle_dependencies"],
        use_random: bool = config["main_engine_configs"]["use_random"],
        use_logger: bool = config["main_engine_configs"]["use_logger"],
        max_depth: int = config["main_engine_configs"]["max_depth"],
        max_breadth: int = config["main_engine_configs"]["max_depth"],
        enable_tracing: bool = config["main_engine_configs"]["enable_tracing"],
        trace_save_directory: str = None,
        database: Database = None,
    ):
        self.mode = "DFS"
        # Passing the common setup to the BaseEngine
        super().__init__(
            headless=headless,
            enable_tracing=enable_tracing,
            trace_save_directory=trace_save_directory,
            database=database,
            use_random=use_random,
            use_logger=use_logger,
            mode=self.mode,
            openai_api_key=openai_api_key,
            vertexai_project_id=vertexai_project_id,
            vertexai_server_location=vertexai_server_location,
            gemini_api_key=gemini_api_key,
        )

        # session_id stays here becasue BaseEngine will be inherited by many
        self.session_id = uuid.uuid4().hex

        selectors = tuple(config["process_config"]["selectors"])
        self.combined_selector = ", ".join(selectors)
        self.planner_agent = PlannerAgent(engine=self)

        self.max_depth = max_depth
        self.max_breadth = max_depth
        self.old_plan = None  # A variable to hold the old plan for the planner agent to understand what has been done already

    async def run(
        self,
        prompt: str,
        automated_login_sites: List[str] = None,
        extraction_format: BaseModel = None,
    ) -> Union[str, None]:
        """
        Run pyba in DFS mode.

        Args:
            `prompt`: The task assigned to DFS by the user
            `automated_login_sites`: Login site name for pre-written scripts to run
            `extraction_format`: A pydantic BaseModel which defines the extraction format for any data extraction

        The task is fed into the planner to get a plan which is then passed to the action models
        to fetch an actionable element.
        """
        if automated_login_sites is not None:
            assert isinstance(
                automated_login_sites, list
            ), "Make sure the automated_login_sites is a list!"

            for engine in automated_login_sites:
                # Each engine is going to be a name like "instagram"
                if hasattr(LoginEngine, engine):
                    engine_class = getattr(LoginEngine, engine)
                    self.automated_login_engine_classes.append(engine_class)
                else:
                    raise UnknownSiteChosen(LoginEngine.available_engines())
        try:
            async with Stealth().use_async(async_playwright()) as p:
                self.browser = await p.chromium.launch(headless=self.headless_mode)

                self.context = await self.get_trace_context()
                self.page = await self.context.new_page()
                cleaned_dom = await initial_page_setup(self.page)

                for steps in range(0, self.max_breadth):
                    # The breadth specifies the number of different plans we can execute
                    plan = self.planner_agent.generate(task=prompt, old_plan=self.old_plan)
                    self.log.info(f"This is the plan for a DFS: {plan}")

                    for _ in range(0, self.max_depth):
                        # The depth is the number of actions for each plan
                        # First check for login
                        login_attempted_successfully = await self.attempt_login()
                        # We'll count logging in as another step in the process
                        if login_attempted_successfully:
                            cleaned_dom = await self.successful_login_clean_and_get_dom()
                            continue
                        # Get an actionable element from the playwright agent

                        # NOTE: This function needs to actually fetch history, but right now its fetching the previous_action only
                        # We need to ensure that we store the previous action REGARDLESS of whether a database has been provided by
                        # the user or not

                        # history = self.fetch_history()
                        previous_action = self.fetch_history()  # This part needs to change...

                        # TODO: This needs to be fixed with the right history implementation
                        action = self.fetch_action(
                            cleaned_dom=cleaned_dom.to_dict(),
                            user_prompt=plan,
                            previous_action=previous_action,
                            extraction_format=extraction_format,
                            action_status=True,  # How do I get the result of the previous action without querying it from the DB?
                            fail_reason=None,
                        )
                        # Check if the automation has finished and if so, get the output
                        output = await self.generate_output(
                            action=action, cleaned_dom=cleaned_dom, prompt=plan
                        )
                        if output:
                            await self.save_trace()
                            await self.shut_down()
                            return output
                        # If not, store the action and perform the action
                        self.log.action(action)

                        value, fail_reason = await perform_action(self.page, action)
                        if value is None:
                            # This means the action failed due to whatever reason. The best bet is to
                            # pass in the latest cleaned_dom and get the output again
                            if self.db_funcs:
                                self.db_funcs.push_to_episodic_memory(
                                    session_id=self.session_id,
                                    action=str(action),
                                    page_url=str(self.page.url),
                                    action_status=False,
                                    fail_reason=fail_reason,
                                )
                            cleaned_dom = await self.extract_dom()
                            output = await self.retry_perform_action(
                                cleaned_dom=cleaned_dom.to_dict(),
                                prompt=plan,
                                previous_action=previous_action,
                                action_status=False,  # This is pretty much unncessary because passing a fail_reason DOES mean that action failed, but it makes it a little more coherent
                                fail_reason=fail_reason,
                            )
                            if output:
                                await self.save_trace()
                                await self.shut_down()
                                return output
                        # Picking the clean DOM now
                        if self.db_funcs:
                            self.db_funcs.push_to_episodic_memory(
                                session_id=self.session_id,
                                action=str(action),
                                page_url=str(self.page.url),
                                action_status=True,
                                fail_reason=None,
                            )
                        cleaned_dom = await self.extract_dom()

                    self.log.warning(
                        "The maximum depth for the current plan has been reached, generating a new plan"
                    )
                    self.old_plan = plan
        finally:
            await self.save_trace()
            await self.shut_down()

    def sync_run(
        self,
        prompt: str,
        automated_login_sites: List[str] = None,
        extraction_format: BaseModel = None,
    ) -> Union[str, None]:
        """
        Sync endpoint for running the above function
        """
        try:
            output = asyncio.run(
                self.run(
                    prompt=prompt,
                    automated_login_sites=automated_login_sites,
                    extraction_format=extraction_format,
                )
            )

            if output:
                return output
        except KeyboardInterrupt:
            # This is a forced shutdown, silently let it slip
            pass



import asyncio
import uuid
from typing import List, Union

from playwright.async_api import async_playwright
from playwright_stealth import Stealth
from pydantic import BaseModel

from pyba.core.agent import PlannerAgent
from pyba.core.lib.action import perform_action
from pyba.core.lib.mode.base import BaseEngine
from pyba.core.scripts import LoginEngine
from pyba.database import Database
from pyba.utils.common import initial_page_setup
from pyba.utils.exceptions import UnknownSiteChosen
from pyba.utils.load_yaml import load_config

config = load_config("general")


class BFS(BaseEngine):
    """
    Methods for handling BFS exploratory searches. The `BaseEngine` initialises
    the provider and with that the playwright action and output agents.

    This is another entry point engine and can be directly imported by the user.

    The following params are defined:

    Args:
        `openai_api_key`: API key for OpenAI models should you want to use that
        `vertexai_project_id`: Create a VertexAI project to use that instead of OpenAI
        `vertexai_server_location`: VertexAI server location
        `gemini_api_key`: API key for Gemini-2.5-pro native support without VertexAI
        `headless`: Choose if you want to run in the headless mode or not
        `handle_dependencies`: Choose if you want to automatically install dependencies during runtime
        `use_logger`: Choose if you want to use the logger (that is enable logging of data)
        `max_depth`: The maximum depth to go into for each plan, where each level of depth corresponds to an action
        `max_breadth`: The number of plans to execute one by one in depth
        `enable_tracing`: Choose if you want to enable tracing. This will create a .zip file which you can use in traceviewer
        `trace_save_directory`: The directory where you want the .zip file to be saved

        `database`: An instance of the Database class which will define all database specific configs

    Find these default values at `pyba/config.yaml`.
    """

    def __init__(
        self,
        openai_api_key: str = None,
        vertexai_project_id: str = None,
        vertexai_server_location: str = None,
        gemini_api_key: str = None,
        headless: bool = config["main_engine_configs"]["headless_mode"],
        handle_dependencies: bool = config["main_engine_configs"]["handle_dependencies"],
        use_logger: bool = config["main_engine_configs"]["use_logger"],
        max_depth: int = config["main_engine_configs"]["max_depth"],
        max_breadth: int = config["main_engine_configs"]["max_depth"],
        enable_tracing: bool = config["main_engine_configs"]["enable_tracing"],
        trace_save_directory: str = None,
        database: Database = None,
    ):
        self.mode = "BFS"
        # Passing the common setup to the BaseEngine
        super().__init__(
            headless=headless,
            enable_tracing=enable_tracing,
            trace_save_directory=trace_save_directory,
            database=database,
            use_logger=use_logger,
            mode=self.mode,
            openai_api_key=openai_api_key,
            vertexai_project_id=vertexai_project_id,
            vertexai_server_location=vertexai_server_location,
            gemini_api_key=gemini_api_key,
        )

        # session_id stays here becasue BaseEngine will be inherited by many
        self.session_id = uuid.uuid4().hex

        selectors = tuple(config["process_config"]["selectors"])
        self.combined_selector = ", ".join(selectors)
        self.planner_agent = PlannerAgent(engine=self)

        self.max_depth = max_depth
        self.max_breadth = max_depth

    # def run(self, task: str):
    # async with Stealth().use_async(async_playwright()) as p:
    #   self.browser = await p.chromium.launch(headless=self.headless_mode)
    #   self.context = await self.get_trace_context()
    #   self.page = await self.context.new_page()
    #   cleaned_dom = await initial_page_setup(self.page)

    #   for steps in range(0, self.max_breadth):
    #       # The breadth specifies the number of different plans we can execute
    #       plan = self.planner_agent.generate(task=task, old_plan=self.old_plan)

    async def _run(
        self, task: str, extraction_format: BaseModel = None, context_id: str = None
    ) -> Union[str, None]:
        """
        helper run function for BFS

        Args:
            `task`: A singular task which needs to be performed
            `extraction_format`: The extraction format for the required goal
            `context_id`: A dynamically generated context-id for each browser window

        Since BFS is going to generated multiple windows at runtime, we assign each one its own ID. This helps manage their
        individual exponential retires and logging.
        """
        try:
            async with Stealth().use_async(async_playwright()) as p:
                browser = await p.chromium.launch(headless=self.headless_mode)

                context = await self.get_trace_context(browser_instance=browser)
                page = await context.new_page()
                cleaned_dom = await initial_page_setup(page)

                for _ in range(0, self.max_depth):
                    # The depth is the number of actions for each plan
                    # First check for login
                    login_attempted_successfully = await self.attempt_login(page)
                    # We'll count logging in as another step in the process
                    if login_attempted_successfully:
                        cleaned_dom = await self.successful_login_clean_and_get_dom()
                        continue
                    # Get an actionable element from the playwright agent
                    previous_action = self.fetch_history()  # We need to fix history implementation for multiple agents in parallel as well!
                    action = self.fetch_action(
                        cleaned_dom=cleaned_dom.to_dict(),
                        user_prompt=task,
                        previous_action=previous_action,
                        extraction_format=extraction_format,
                        context_id=context_id,
                        action_status=True,
                        fail_reason=None,
                    )
                    # Check if the automation has finished and if so, get the output
                    output = await self.generate_output(
                        action=action, cleaned_dom=cleaned_dom, prompt=task
                    )
                    if output:
                        await self.save_trace()
                        await self.shut_down()
                        return output
                    # If not, perform the action
                    self.log.action(action)

                    value, fail_reason = await perform_action(page, action)
                    if value is None:
                        # This means the action failed due to whatever reason. The best bet is to
                        # pass in the latest cleaned_dom and get the output again
                        if self.db_funcs:
                            self.db_funcs.push_to_bfs_episodic_memory(
                                session_id=self.session_id,
                                context_id=context_id,
                                action=str(action),
                                page_url=str(page.url),
                            )
                        cleaned_dom = await self.extract_dom(page)
                        output = await self.retry_perform_action(
                            cleaned_dom=cleaned_dom.to_dict(),
                            prompt=task,
                            previous_action=previous_action,
                            action_status=False,
                            fail_reason=fail_reason,
                            page=page,
                        )
                        if output:
                            await self.save_trace(context)
                            await self.shut_down(context, browser)
                            return output
                        # Continue to the next iteration after retry
                        continue

                    # Action succeeded, log and get the new DOM
                    if self.db_funcs:
                        self.db_funcs.push_to_bfs_episodic_memory(
                            session_id=self.session_id,
                            context_id=context_id,
                            action=str(action),
                            page_url=str(page.url),
                        )
                    cleaned_dom = await self.extract_dom(page)

                self.log.warning(
                    "The maximum depth for the current task has been reached, generating a new plan to achieve this task"
                )
        finally:
            await self.save_trace(context)
            await self.shut_down(context, browser)

    async def run(
        self,
        prompt: str,
        automated_login_sites: List[str] = None,
        extraction_format: BaseModel = None,
    ) -> List:
        """
        The async run function

        Args:
            `prompt`: The prompt which needs to be converted to plans
            `automated_login_sites`: List of names for which sites to login automatically
            `extraction_format`: The extraction format for any extraction that needs to be done

        Returns:
            List
        """

        if automated_login_sites is not None:
            assert isinstance(
                automated_login_sites, list
            ), "Make sure the automated_login_sites is a list!"

            for engine in automated_login_sites:
                # Each engine is going to be a name like "instagram"
                if hasattr(LoginEngine, engine):
                    engine_class = getattr(LoginEngine, engine)
                    self.automated_login_engine_classes.append(engine_class)
                else:
                    raise UnknownSiteChosen(LoginEngine.available_engines())

        plan_list = self.planner_agent.generate(task=prompt)

        print(plan_list)
        assert isinstance(
            plan_list, list
        ), f"Expected the plan to be a list, got {type(plan_list)} instead."

        self.log.info(f"This is the plan for a BFS: {plan_list}")

        # Keeping this purely async is better for playwright
        tasks = []
        for task in plan_list:
            context_id = uuid.uuid4().hex  # This should do it.
            tasks.append(asyncio.create_task(self._run(task, extraction_format, context_id)))
        results = await asyncio.gather(*tasks, return_exceptions=False)

        return results

    def sync_run(
        self,
        prompt: str,
        automated_login_sites: List[str] = None,
        extraction_format: BaseModel = None,
    ):
        """
        Sync endpoint for running the BFS mode (to be used from synchrnous code)

        Args:
            `prompt`: The prompt which needs to be converted to plans
            `automated_login_sites`: List of names for which sites to login automatically
            `extraction_format`: The extraction format for any extraction that needs to be done
        """
        try:
            output = asyncio.run(
                self.run(
                    prompt=prompt,
                    automated_login_sites=automated_login_sites,
                    extraction_format=extraction_format,
                )
            )

            if output:
                return output
        except KeyboardInterrupt:
            # This is a forced shutdown, silently let it slip
            pass



system_prompt = """
You are the **Brain** of a browser automation engine.

Your purpose is to reason about a web page’s current state and produce the **next precise, atomic instruction** for a Playwright automation script.

---

## What You Receive

You are given two inputs:

1. **Task Description (User Goal):**
   - A clear instruction of what needs to be achieved (e.g., "Search for a product and add it to the cart").
   - This represents the *next logical intent* in the browsing flow.

2. **Cleaned DOM (Context of Current Page):**
   A structured dictionary extracted from the current web page, containing:
   - `hyperlinks`: list of all hyperlink texts or targets.
   - `input_fields`: list of all fillable input elements.
   - `clickable_fields`: list of clickable elements (buttons, spans, etc.).
   - `actual_text`: visible text content of the page.

Use this cleaned DOM to understand what’s visible, available, and interactable on the page right now.

---

## What You Must Output

You must **always output a valid JSON object** of type `PlaywrightResponse`, defined as:

```python
class PlaywrightResponse(BaseModel):
    actions: List[PlaywrightAction]
    extract_info: bool
```

Each PlaywrightAction represents one atomic browser operation (click, type, fill, press, scroll, etc.).
All actions must strictly follow the structure and semantics of the fields below.

## PlaywrightAction Schema

Each field represents a possible browser action.
You must set `exactly one atomic action` (or one valid pair like fill_selector + fill_value) per output.
All other fields should remain null or absent.

### Navigation

- goto: Navigate to a given URL using page.goto(url).
- go_back: Go back in browser history using page.go_back().
- go_forward: Go forward in browser history using page.go_forward().
- reload: Reload the current page using page.reload().

### Interactions

- click: Click the specified element using page.click(selector).
- dblclick: Double-click the specified element using page.dblclick(selector).
- hover: Hover over the specified element using page.hover(selector).
- right_click: Right click a specified element using page.click(selector, button='right')

### Input Actions

- fill_selector: Selector of the input element to fill with text using page.fill().
- fill_value: Text value to insert into the element specified by fill_selector.
- type_selector: Selector of the element to type into using page.type().
- type_text: Text to type into the element specified by type_selector.
- press_selector: Selector of the element to send a key press event to using page.press().
- press_key: Key to press (e.g., "Enter", "Escape", "ArrowDown") on the element in press_selector.
- check: Selector of a checkbox or radio button to check using page.check().
- uncheck: Selector of a checkbox or radio button to uncheck using page.uncheck().
- select_selector: Selector of a <select> dropdown element to modify using page.select_option().
- select_value: Option value to select in the element specified by select_selector.
- upload_selector: Selector of a file input element for upload using page.set_input_files().
- upload_path: Path to the file(s) to upload for upload_selector.

### Scrolling & Waiting

- scroll_x: Horizontal scroll position to move to using page.evaluate().
- scroll_y: Vertical scroll position to move to using page.evaluate().
- wait_selector: Selector to wait for before continuing, using page.wait_for_selector().
- wait_timeout: Maximum time (in ms) to wait for the selector or event.
- wait_ms: Wait for a fixed duration (in milliseconds) using time.sleep().

## Clicking on dropdown menus (**you must specify both of these together**)

- dropdown_field_id: A selector for the dropdown field
- dropdown_field_value: The value that needs to be chosen from the dropdown menu

### Keyboard & Mouse

- keyboard_press: Key to simulate pressing using page.keyboard.press().
- keyboard_type: Text to type using page.keyboard.type().
- mouse_move_x: X-coordinate to move the mouse to using page.mouse.move().
- mouse_move_y: Y-coordinate to move the mouse to using page.mouse.move().
- mouse_click_x: X-coordinate for a direct mouse click using page.mouse.click().
- mouse_click_y: Y-coordinate for a direct mouse click using page.mouse.click().

### Page & Context Management

- new_page: Create a new browser page (optionally open a URL).
- close_page: Close the current page using page.close().
- switch_page_index: Switch to another open page by its index in context.pages().

### Evaluation & Utilities

- evaluate_js: Run custom JavaScript in the browser context using page.evaluate(script).
- screenshot_path: Path to save a screenshot using page.screenshot().
- download_selector: Selector to trigger a file download from (e.g., a link or button).

## Rules for Output

### Atomicity

- You must produce exactly one atomic PlaywrightAction per response.
- Only one action (or a single valid pair like fill_selector + fill_value) may be non-null.
- All other fields must remain null or be omitted.
- Complex operations (like filling and pressing Enter) must be split into separate sequential responses.

### Sequentiality

- Each response represents the next step in the browsing sequence.
- Multi-step actions (like search → click result) must unfold step-by-step across responses.

### Contextual Validity

- You may only reference elements that exist in the provided cleaned_dom.
- Do not invent or assume selectors or page content.

### Intent Awareness

- Match the user’s goal with what’s currently visible and available.
- If you just filled a relevant input field, the next step may be pressing Enter.
- If no suitable element exists, choose the most relevant visible field and trigger a generic completion action (like pressing Enter).

### Extract Info Condition

- Set extract_info to true if the current page visibly contains information relevant to the user goal.
- Otherwise set it to false.

NOTE: IF THE USER HAS REQUESTED FOR CERTAIN EXTRACTIONS, DON'T TRY TO DO IT YOURSELF. SET THE `extract_info` BOOLEAN TO TRUE AND PROCEED (OR SET A WAIT TIME IN ACTIONS)

### Completion Condition

If the task appears complete, or no further actions can be taken, output: `None`

## Allowed Examples

- Example 1 — Filling a Search Box

{{
  "actions": [
    {{
      "fill_selector": "input[name='q']",
      "fill_value": "Playwright Python"
    }}
  ],
  "extract_info": false,
}}


- Example 2 — Pressing Enter on the Same Box

{{
  "actions": [
    {{
      "press_selector": "input[name='q']",
      "press_key": "Enter"
    }}
  ],
  "extract_info": true,
}}


Note: Your choice for the `extract_info` must only be decided based on what is available on the current page.

## Disallowed Examples

- Multiple actions in one step (invalid):

{{
  "actions": [
    {{
      "fill_selector": "input[name='q']",
      "fill_value": "Playwright",
      "press_selector": "input[name='q']",
      "press_key": "Enter"
    }}
  ], 
  "extract_info": false
}}


## Summary

Think like a cautious human tester:

- Perform one reliable action at a time.
- Never guess selectors.
- Stay consistent with the user’s goal.
- Always produce structured JSON.
- When the task is done or unclear — output None.
- Always include extract_info (true or false).

NOTE: IF THE USER HAS REQUESTED FOR CERTAIN EXTRACTIONS, DON'T TRY TO DO IT YOURSELF. SET THE `extract_info` BOOLEAN TO TRUE AND PROCEED (OR SET A WAIT TIME IN ACTIONS)
"""



import asyncio
import json
from typing import Dict, Optional, Literal

from playwright.async_api import TimeoutError
from pydantic import BaseModel

import pyba.core.helpers as global_vars
from pyba.core.agent import PlaywrightAgent
from pyba.core.helpers.jitters import MouseMovements, ScrollMovements
from pyba.core.lib import HandleDependencies
from pyba.core.lib.action import perform_action
from pyba.core.lib.code_generation import CodeGeneration
from pyba.core.provider import Provider
from pyba.core.scripts import ExtractionEngines
from pyba.core.tracing import Tracing
from pyba.database import DatabaseFunctions
from pyba.logger import setup_logger, get_logger
from pyba.utils.exceptions import DatabaseNotInitialised


class BaseEngine:
    """
    A reusable base class that encapsulates the shared browser lifecycle,
    tracing, DOM extraction, and utility helpers.

        The following will be initialised by the BaseEngine:

        - `db_funcs`: The database functions to be used for inserting and querying logs
        - `mode`: The mode of operation (DFS, BFS or Normal)
        - `provider_instance`: This will detect the provider you're using
        - `playwright_agent`: The actual playwright brains of the operation
    """

    def __init__(
        self,
        headless: bool = True,
        enable_tracing: bool = True,
        trace_save_directory: str = None,
        database=None,
        use_random=None,
        use_logger: bool = None,
        mode: Literal["DFS", "BFS", "Normal", "STEP"] = None,
        handle_dependencies: bool = False,
        openai_api_key: str = None,
        vertexai_project_id: str = None,
        vertexai_server_location: str = None,
        gemini_api_key: str = None,
    ):
        self.headless_mode = headless
        self.tracing = enable_tracing
        self.trace_save_directory = trace_save_directory

        self.mode = mode
        self.database = database
        self.db_funcs = DatabaseFunctions(self.database) if database else None

        self.automated_login_engine_classes = []

        self.use_random_flag = (
            use_random if use_random else False
        )  # I like to set defaults as None...
        global_vars._use_random = (
            self.use_random_flag
        )  # Update the global use random for other modules

        setup_logger(use_logger=use_logger)
        self.log = get_logger()

        provider_instance = Provider(
            openai_api_key=openai_api_key,
            gemini_api_key=gemini_api_key,
            vertexai_project_id=vertexai_project_id,
            vertexai_server_location=vertexai_server_location,
        )

        self.provider = provider_instance.provider
        self.model = provider_instance.model
        self.openai_api_key = provider_instance.openai_api_key
        self.gemini_api_key = provider_instance.gemini_api_key
        self.vertexai_project_id = provider_instance.vertexai_project_id
        self.location = provider_instance.location

        # Defining the playwright agent with the defined configs
        self.playwright_agent = PlaywrightAgent(engine=self)

        if handle_dependencies:
            HandleDependencies.playwright.handle_dependencies()

    async def run(self):
        """
        Run function which will be defined inside all child classes
        """
        pass

    async def extract_dom(self, page=None):
        """
        Extracts the relevant fields from the DOM of the current page and returns
        the DOM dataclass. This is backwards compatible with Engine and DFS while
        it supports BFS by pinning the page down.

        Args:
            `page`: Optional argument to pin the page for removing self dependency
        """
        page_obj = page if page is not None else self.page
        self.mouse = MouseMovements(page=page_obj)
        self.scroll_manager = ScrollMovements(page=page_obj)

        try:
            await self.wait_till_loaded(page_obj)
            page_html = await page_obj.content()
        except Exception:
            # We might get a "Unable to retrieve content because the page is navigating and changing the content" exception
            # This might happen because page.content() will start and issue an evaluate, while the page is reloading and making network calls
            # So, once it gets a response, it commits it and clears the execution contents so page.content() fails.
            # See https://github.com/microsoft/playwright/issues/16108

            # We might choose to wait for networkidle -> https://github.com/microsoft/playwright/issues/22897
            try:
                await self.wait_till_loaded(page_obj)
            except Exception:
                # If networkidle never happens, then we'll try a direct wait
                await asyncio.sleep(3)

            page_html = await page_obj.content()

        try:
            body_text = await page_obj.inner_text("body")
            elements = await page_obj.query_selector_all(self.combined_selector)
            base_url = page_obj.url
        except TimeoutError:
            self.log.error("The page has not loaded within the defined timeout, going back")
            return None

        # Then we need to extract the new cleaned_dom from the page
        # Passing in known_fields for the input fields that we already know off so that
        # its easier for the extraction engine to work
        extraction_engine = ExtractionEngines(
            html=page_html,
            body_text=body_text,
            elements=elements,
            base_url=base_url,
            page=page_obj,
        )

        # Perform an all out extraction
        cleaned_dom = await extraction_engine.extract_all()
        cleaned_dom.current_url = base_url
        return cleaned_dom

    async def generate_output(self, action, cleaned_dom, prompt):
        """
        Helper function to generate the output if the action
        has been completed.

        Args:
            `action`: The action as given out by the model
            `cleaned_dom`: The latest cleaned_dom for the model to read
            `prompt`: The prompt which was given to the model
        """
        if action is None or all(value is None for value in vars(action).values()):
            self.log.success("Automation completed, agent has returned None")
            try:
                output = self.playwright_agent.get_output(
                    cleaned_dom=cleaned_dom.to_dict(), user_prompt=prompt
                )
                self.log.info(f"This is the output given by the model: {output}")
                return output
            except Exception:
                # This should rarely happen
                await asyncio.sleep(10)
                output = self.playwright_agent.get_output(
                    cleaned_dom=cleaned_dom.to_dict(), user_prompt=prompt
                )
                self.log.info(f"This is the output given by the model: {output}")
                return output
        else:
            return None

    async def save_trace(self, context=None):
        """
        Endpoint to save the trace if required. This is backwards compatible for Engine
        and DFS while is supports BFS by removing the self dependency

        Args:
            `context`: Optional argument to pin the browser context down
        """
        context_obj = context if context is not None else self.context
        if self.tracing:
            trace_path = self.trace_dir / f"{self.session_id}_trace.zip"
            try:
                await context_obj.tracing.stop(path=str(trace_path))
                self.log.info(f"This is the tracepath: {trace_path}")
            except Exception:
                # Abrupt browser closure
                pass

    async def shut_down(self, context=None, browser=None):
        """
        Function to cleanly close the existing browsers and contexts. This also saves
        the traces in the provided trace_dir by the user or the default.

        This is backwards compatible for Engine and DFS while is supports BFS by
        removing the self dependency

        Args:
            `context`: Optional argument to pin the specific brower context down
            `browser`: Optional argument to pin the browser instance down
        """
        context_obj = context if context is not None else self.context
        browser_obj = browser if browser is not None else self.browser
        try:
            await context_obj.close()
            await browser_obj.close()
        except Exception:
            # Context/browser have already been closed
            pass

    def generate_code(self, output_path: str) -> bool:
        """
        Function end-point for code generation

        Args:
            `output_path`: output file path to save the generated code to
        """
        if not self.db_funcs:
            raise DatabaseNotInitialised()

        codegen = CodeGeneration(
            session_id=self.session_id, output_path=output_path, database_funcs=self.db_funcs
        )
        codegen.generate_script()
        self.log.info(f"Created the script at: {output_path}")
        return True

    async def get_trace_context(self, browser_instance=None):
        """
        Helper function to intialise the context using the Tracing class. This is backwards compatible
        for Engine and DFS while is supports BFS by removing the self dependency

        Args:
            `browser_instance`: Optional argument to pin the browser session down

        Return:
            `context`: The playwright to be used for automation
        """

        tracing = Tracing(
            browser_instance=browser_instance if browser_instance is not None else self.browser,
            session_id=self.session_id,
            enable_tracing=self.tracing,
            trace_save_directory=self.trace_save_directory,
        )

        self.trace_dir = tracing.trace_dir
        context = await tracing.initialize_context()

        return context

    async def attempt_login(self, page=None) -> bool:
        """
        Helper function to attempt and perform a login to chosen sites. This is backwards compatible
        with Engine and DFS while it supports BFS by pinning the page down.

        Args:
            `page`: Optional argument to pin the page for removing self dependency

        Returns:
            `flag`: A boolean to indicate the success or failure for the attempt

        The login attempt may fail due to two reasons:

            - The current page is not a login page
            - Some selectors changed due to which the login engine returned None

        Note that the LoginEngines are hardcoded engines for speed.
        """

        flag = False
        page_obj = page if page is not None else self.page
        if self.automated_login_engine_classes:
            for engine in self.automated_login_engine_classes:
                engine_instance = engine(page_obj)
                self.log.info(f"Testing for {engine_instance.engine_name} login engine")
                # Instead of just running it and checking inside, we can have a simple lookup list
                out_flag = await engine_instance.run()
                if out_flag:
                    # This means it was True and we successfully logged in
                    self.log.success(f"Logged in successfully through the {page_obj.url} link")
                    flag = True
                    break
                elif out_flag is None:
                    # This means it wasn't for a login page for this engine
                    pass
                else:
                    # This means it failed
                    self.log.warning(f"Login attempted at {page_obj.url} but failed!")

        return flag

    async def successful_login_clean_and_get_dom(self, page=None):
        """
        Helper function to obtain the cleaned_dom after a successful login. This is backwards compatible
        with Engine and DFS while it supports BFS by pinning the page down.

        Args:
            `page`: Optional argument to pin the page for removing self dependency

        Functionality:

        - Cleans the automated_login_engine_classes list (that is, we're assuming only 1 login session
        for each run)
        - Gets the latest page contents and parses the DOM using the extraction engine
        """
        page_obj = page if page is not None else self.page

        self.automated_login_engine_classes = None
        self.mouse = MouseMovements(page=page_obj)
        self.scroll_manager = ScrollMovements(page=page_obj)
        # Update the DOM after a login
        try:
            await self.wait_till_loaded()
        except Exception:
            await asyncio.sleep(2)

        page_html = await page_obj.content()
        body_text = await page_obj.inner_text("body")
        elements = await page_obj.query_selector_all(self.combined_selector)
        base_url = page_obj.url

        extraction_engine = ExtractionEngines(
            html=page_html,
            body_text=body_text,
            elements=elements,
            base_url=base_url,
            page=page_obj,
        )
        cleaned_dom = await extraction_engine.extract_all()
        cleaned_dom.current_url = base_url

        return cleaned_dom

    def fetch_history(self) -> str:
        """
        Helper function to obtain the history of actions.

        TODO: This functions should fetch the last k history elements and use them as `history` and
        NOT previous_action. The previous_action and its status must be stored regardless.

        Returns:
            `history`: The last logged action
        """

        try:
            # Get history if db_funs is defined, that is, Databases are being used
            history = None
            if self.db_funcs:
                history = self.db_funcs.get_episodic_memory_by_session_id(
                    session_id=self.session_id
                )

            history = json.loads(history.actions)[-1] if history else ""
        except Exception as e:
            self.log.warning(f"Couldn't query the database for history: {e}")
            history = ""

        return history

    def fetch_action(
        self,
        cleaned_dom: Dict,
        user_prompt: str,
        previous_action: str = None,
        extraction_format: BaseModel = None,
        context_id: str = None,
        fail_reason: str = None,
        action_status: bool = None,
    ):
        """
        Helper function to fetch an actionable PlaywrightResponse element

        Args:
            `cleaned_dom`: The DOM for the current page
            `user_prompt`: The actual task given by the user
            `previous_action`: The last action performed by the model
            `extraction_format`: The extraction format requested by the user.
            `context_id`: A unique identifier for this browser window (useful when multiple windows)
            `fail_reason`: The reason for the failure of the previous action
            `action_status`: A boolean to indicate if the previous action was successful or not

        For an explanation of the `extraction_format` read the main file documentation.

        Returns:
            `action`: An actionable playwrightresponse element
        """

        try:
            # Each process-action call is to be passed the status of the previous call
            action = self.playwright_agent.process_action(
                cleaned_dom=cleaned_dom,
                user_prompt=user_prompt,
                previous_action=previous_action,
                extraction_format=extraction_format,
                context_id=context_id,
                fail_reason=fail_reason,
                action_status=action_status,
            )
        except Exception as e:
            self.log.error(f"something went wrong in obtaining the response: {e}")
            action = None

        return action

    async def retry_perform_action(
        self,
        cleaned_dom: Dict,
        prompt: str,
        previous_action: str,
        action_status: bool,
        fail_reason: str,
        extraction_format: BaseModel = None,
        page=None,
    ) -> Optional[str]:
        """
        helper function to retry the action after a failure. This is backwards compatible with Engine
        and DFS while it supports BFS by pinning the page down.


        Args:
            `cleaned_dom`: The new cleaned DOM for the current page
            `prompt`: The original prompt given by the user
            `previous_action`: The past action that failed
            `action_status`: A boolean to indicate the failure of the action (I know, not needed but let's keep it for now!)
            `fail_reason`: Reason for the failure for the action
            `extraction_format`: In case the current page needs extraction as well
            `page`: Optional argument to pin the page down to remove self dependency

        This function will retry the action based on the current DOM and the past action. This should
        most likely fix the issue of a stale element or a hallucinated component or something.

        Return:
            `output`: If the action was successful and automation is completed
            `None`: The usual case where an action is performed
        """
        page_obj = page if page is not None else self.page

        self.log.warning("The previous action failed, checking the latest page")
        action = self.playwright_agent.process_action(
            cleaned_dom=cleaned_dom,
            user_prompt=prompt,
            previous_action=previous_action,
            fail_reason=fail_reason,
            extraction_format=extraction_format,
            action_status=action_status,
        )

        output = await self.generate_output(action=action, cleaned_dom=cleaned_dom, prompt=prompt)

        if output:
            return output

        self.log.action(action)

        # Deprecated. We now log data before calling retry action
        # if self.db_funcs:
        #     self.db_funcs.push_to_episodic_memory(
        #         session_id=self.session_id,
        #         action=str(action),
        #         page_url=str(page_obj.url),
        #     )

        await perform_action(page_obj, action)

    async def wait_till_loaded(self, page=None):
        """
        Helper function to wait till load state while applying random jitters
        (if specified by the user). This is backwards compatible with Engine
        and DFS while it supports BFS by pinning the page down.

        Args:
            `page`: Optional argument to pin the page for removing self dependency
        """
        page_obj = page if page is not None else self.page
        if self.use_random_flag:
            await asyncio.gather(
                page_obj.wait_for_load_state("networkidle", timeout=1000),
                self.mouse.random_movement(),
                self.scroll_manager.apply_scroll_jitters(),
            )  # Wait for a second for network calls to stablize
        else:
            (await page_obj.wait_for_load_state("networkidle", timeout=1000),)



extraction_system_instruction = """
You are the extraction agent.

You receive the full text of the current screen, which may contain irrelevant junk.
You also receive user instructions and a target output schema.
Your job is to extract only the information explicitly requested by the user and fill the output exactly according to the schema.

# Important
Do not invent or infer anything that is not present in the provided text. If required fields cannot be filled, use a clear placeholder such as “nothing available”.
Keep the output strictly valid to the specified format and provide no extra commentary.
"""

extraction_general_instruction = """
You are the extraction agent.

You receive the full text of the current screen, which may contain irrelevant junk.
You also receive user instructions and a target output schema.
Your job is to extract only the information explicitly requested by the user and fill the output exactly according to the schema.

This is the user's request:

{task}

This is the actual text on the current page:

{actual_text}
"""
